@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2011-2013 Canonical
+ * Copyright (C) 2011-2014 Canonical
  * Hugely modified parts from powertop-1.13, Copyright 2007, Intel Corporation
  *
  * This program is free software; you can redistribute it and/or
@@ -45,6 +45,9 @@
 #define OPT_DIRNAME_STRIP	(0x00000010)
 #define OPT_SAMPLE_COUNT	(0x00000020)
 #define OPT_RESULT_STATS	(0x00000040)
+#define OPT_BRIEF		(0x00000080)
+#define OPT_KERNEL		(0x00000100)
+#define OPT_USER		(0x00000200)
 #define OPT_CMD			(OPT_CMD_SHORT | OPT_CMD_LONG)
 
 typedef struct link {
@@ -407,26 +410,16 @@ static int info_compare_total(const void *item1, const void *item2)
  */
 static bool pid_a_kernel_thread(const char *task, const pid_t id)
 {
-	char buffer[128];
-	char path[128];
-
 	if (sane_procs) {
-		int piddir;
-		bool ret = false;
-
-		snprintf(buffer, sizeof(buffer), "/proc/%d", id);
-		if ((piddir = openat(0, buffer, O_RDONLY)) < 0)
-			return ret;	/* it's no more, who knows what it was */
-		
-		if (readlinkat(piddir, "exe", path, sizeof(path)) < 0) 
-			if (errno == ENOENT)
-				ret = true;
-		
-		close(piddir);
-		return ret;
+		return getpgid(id) == 0;
 	} else {
 		/* In side a container, make a guess at kernel threads */
 		int i;
+		pid_t pgid = getpgid(id);
+
+		/* This fails for kernel threads inside a container */
+		if (pgid >= 0)
+			return pgid == 0;
 
 		/*
 		 * This is not exactly accurate, but if we can't look up
@@ -751,9 +744,10 @@ static void timer_info_list_free(void)
  */
 static unsigned long hash_pjw(const char *str)
 {
-  	unsigned long h=0, g;
+  	unsigned long h = 0;
 
 	while (*str) {
+		unsigned long g;
 		h = (h << 4) + (*str);
 		if (0 != (g = h&0xf0000000)) {
 			h = h ^ (g >> 24);
@@ -902,9 +896,6 @@ static void timer_stat_diff(
 	timer_stat_t *timer_stats_new[])/* new timer stats samples */
 {
 	int i;
-	int j = 0;
-	unsigned long total = 0UL;
-	unsigned long kt_total = 0UL;
 	double dur = timeval_double(duration);
 
 	timer_stat_t *sorted = NULL;
@@ -933,21 +924,35 @@ static void timer_stat_diff(
 	}
 
 	if (!(opt_flags & OPT_QUIET)) {
-		printf("%8s %-5s %-15s %-25s %-s\n",
-			opt_flags & OPT_CUMULATIVE ? "Events" : "Event/s",
-			"PID", "Task", "Init Function", "Callback");
+		unsigned long total = 0UL, kt_total = 0UL;
+		int j = 0;
+
+		printf("%8s %-5s %-15s",
+			opt_flags & OPT_CUMULATIVE ? "Events" : "Event/s", "PID", "Task");
+		if (!(opt_flags & OPT_BRIEF))
+			printf(" %-25s %-s\n",
+				"Init Function", "Callback");
+		else
+			printf("\n");
 
 		while (sorted) {
 			if (((n_lines == -1) || (j < n_lines)) && (sorted->delta != 0)) {
 				j++;
-				if (opt_flags & OPT_CUMULATIVE) {
-					printf("%8lu %5d %-15s %-25s %-s\n",
-						sorted->count,
-						sorted->info->pid, sorted->info->task,
-						sorted->info->func, sorted->info->callback);
+				if (opt_flags & OPT_CUMULATIVE)
+					printf("%8lu ", sorted->count);
+				else
+					printf("%8.2f ", (double)sorted->delta / dur);
+
+				if (opt_flags & OPT_BRIEF) {
+					char *cmd = sorted->info->cmdline ?
+						sorted->info->cmdline : sorted->info->task;
+
+					printf("%5d %s\n",
+						sorted->info->pid,
+						opt_flags & OPT_CMD ?
+							cmd : sorted->info->task);
 				} else {
-					printf("%8.2f %5d %-15s %-25s %-s\n",
-						(double)sorted->delta / dur,
+					printf("%5d %-15s %-25s %-s\n",
 						sorted->info->pid, sorted->info->task,
 						sorted->info->func, sorted->info->callback);
 				}
@@ -992,6 +997,7 @@ static void get_events(timer_stat_t *timer_stats[])	/* hash table to populate */
 		char func[128];
 		char timer[128];
 		bool kernel_thread;
+		int mask;
 
 		if (fgets(buf, sizeof(buf), fp) == NULL)
 			break;
@@ -1014,8 +1020,11 @@ static void get_events(timer_stat_t *timer_stats[])	/* hash table to populate */
 			continue;	/* Deferred event, skip */
 
 		ptr++;
-		sscanf(buf, "%lu", &count);
-		if (sscanf(ptr, "%d %s %s (%[^)])", &pid, task, func, timer) != 4)
+		sscanf(buf, "%21lu", &count);
+		memset(task, 0, sizeof(task));
+		memset(func, 0, sizeof(func));
+		memset(timer, 0, sizeof(timer));
+		if (sscanf(ptr, "%10d %63s %127s (%127[^)])", &pid, task, func, timer) != 4)
 			continue;
 
 		kernel_thread = pid_a_kernel_thread(task, pid);
@@ -1024,6 +1033,11 @@ static void get_events(timer_stat_t *timer_stats[])	/* hash table to populate */
 		if (strncmp(task, "swapper", 6) == 0)
 			kernel_thread = true;
 
+		mask = kernel_thread ? OPT_KERNEL : OPT_USER;
+
+		if (!(opt_flags & mask))
+			continue;
+
 		if (kernel_thread) {
 			char tmp[64];
 			task[13] = '\0';
@@ -1082,10 +1096,13 @@ int main(int argc, char **argv)
 	list_init(&sample_list);
 
 	for (;;) {
-		int c = getopt(argc, argv, "cCdsSlhn:qr:t:");
+		int c = getopt(argc, argv, "bcCdksSlhn:qr:t:u");
 		if (c == -1)
 			break;
 		switch (c) {
+		case 'b':
+			opt_flags |= OPT_BRIEF;
+			break;
 		case 'c':
 			opt_flags |= OPT_CUMULATIVE;
 			break;
@@ -1127,9 +1144,18 @@ int main(int argc, char **argv)
 		case 'l':
 			opt_flags |= OPT_CMD_LONG;
 			break;
+		case 'k':
+			opt_flags |= OPT_KERNEL;
+			break;
+		case 'u':
+			opt_flags |= OPT_USER;
+			break;
 		}
 	}
 
+	if (!(opt_flags & (OPT_KERNEL | OPT_USER)))
+		opt_flags |= (OPT_KERNEL | OPT_USER);
+
 	if (optind < argc) {
 		duration_secs = atof(argv[optind++]);
 		if (duration_secs < 0.5) {
